/* ----------------------------------------------------------------------------
 *
 * Raima Database Manager
 *
 * Copyright (c) 2012 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 *
 * ----------------------------------------------------------------------------
 */
#include <process.h>
#include <aclapi.h>
#include <stdio.h>

/* rdm.h is required in all files using the RDM Database */
#include "rdm.h"

/* core14_db.h was generated from the DDL file core14_db.ddl by the ddlp.
 * utility.
 */
#include "core14_db.h"

/* core14_db_dbd.h was generated from the DDL
 * file core14_db.ddl by the ddlp utility.  It is required to use the d_open_ptr API
 */
#include "core14_db_dbd.h"

static uint8_t quit = 0;

/* EXAMPLE - Core for C */

/* This example shows multi-user access to a database using separate threads.
 * Two writer threads are created.  The first begins inserting values into
 * the table.  The second waits until the table contains more than 100 rows
 * and then begins deleting the oldest.  A reader thread is also created
 * that prints out the total number of records, the max value, and the min
 * value.  Each thread uses regular locking.
 *
 * The DDL (Database Definition Language) is in core14_db.ddl. The database
 * schema is embedded into the application and the database will be created 
 * the first time the application is run.  Subsequent runs will use the now
 * existing database
 *
 * The following RDM Functions are illustrated:
 *      d_opentask/d_closetask
 *      d_open_ptr/d_close
 *      d_initialize
 *      d_reclock/d_recfree
 *      d_trbegin/d_trend
 *      d_fillnew
 *      d_delete
 *      d_rectot
 *      d_keyfrst/d_keynext
 *      d_recread
 *      d_errorname
 *      d_errorstr
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 *
 * Each of these functions returns an integer status code, where the
 * value S_OKAY indicates a successful call.
 *
 * For simplicity, this example does not check all return codes, but good
 * programming practices would dictate that they are checked after each
 * RDM call.
 */

/* Inserter thread */
static uint32_t __stdcall InserterThreadFcn(
    void *arg)
{
    DB_TASK      *task;
    uint32_t      ii;
    int32_t       rdm_status;
    struct rec01  rec01_rec;

    /* Create a task for this thread (each threads needs to use a private task
     * unless there is some way to ensure that only one thread can use a task
     * at a time.  Multiple threads using the same RDM task simultaneously
     * can/will result in application crashes and database corruption
     */
    rdm_status = d_opentask(&task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error open DB_TASK in InserterThreadFcn - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        _endthreadex( rdm_status );
        return rdm_status;
    }

    /* Open the database in shared mode */
    rdm_status = d_open("core14_db", "s", task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error opening database in InserterThreadFcn - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        d_closetask(task);
        
        _endthreadex( rdm_status );
        return rdm_status;
    }

    /* Loop until we get an error or are told to quit */
    for (ii = 0; !quit && rdm_status == S_OKAY; ii++)
    {
        /* Start a transaction */
        rdm_status = d_trbegin("AddRecord", task);
        if (rdm_status == S_OKAY)
        {
            /* Lock the REC01 record */
            rdm_status = d_reclock(REC01, "w", task, CURR_DB);
            if (rdm_status == S_OKAY)
            {
                /* Add a new record instance */
                rec01_rec.field01 = ii;
                rdm_status = d_fillnew(REC01, &rec01_rec, task, CURR_DB);
                
                /* Commit or rollback the tranaction */
                if (rdm_status == S_OKAY)
                {
                    d_trend(task);
                }
                else
                {
                    d_trabort(task);
                }
            }
        }
    }

    /* Close the database and task */
    d_close(task);
    d_closetask(task);
    
    _endthreadex( 0 );
    return 0;
}

static uint32_t __stdcall DeleterThreadFcn(
    void *arg)
{
    DB_TASK      *task;
    int32_t       rdm_status;
    uint64_t      rec_count = 0;

    /* Create a task for this thread */
    rdm_status = d_opentask(&task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error open DB_TASK in DeleterThreadFcn - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        
        _endthreadex( rdm_status );
        return rdm_status;
    }

    /* Open the database in shared mode */
    rdm_status = d_open("core14_db", "s", task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error opening database in DeleterThreadFcn - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        d_closetask(task);
        
        _endthreadex( rdm_status );
        return rdm_status;
    }

    /* Loop until we get an error or are told to quit */
    while (!quit && rdm_status == S_OKAY)
    {
        /* Check the total number of REC01 record instances */
        rdm_status = d_reclock(REC01, "r", task, CURR_DB);
        if(rdm_status == S_OKAY)
        {
            rdm_status = d_rectot(REC01, &rec_count, task, CURR_DB);
            d_recfree(REC01, task, CURR_DB);
        }

        /* When we have more than 100 records start deleting the record with
         * the smallest FIELD01 value
         */
        if( rdm_status == S_OKAY && rec_count > 100)
        {
            /* Start a transaction */
            rdm_status = d_trbegin("DeleteRecord", task);
            if (rdm_status == S_OKAY)
            {
                /* Lock the REC01 record */
                rdm_status = d_reclock(REC01, "w", task, CURR_DB);
                if (rdm_status == S_OKAY)
                {
                    /* Navigate to the smallest record */
                    rdm_status = d_keyfrst(REC01_FIELD01, task, CURR_DB);
                    if(rdm_status == S_OKAY)
                    {
                        /* Delete the record instance */
                        rdm_status = d_delete(task, CURR_DB);
                
                        /* Commit or rollback the tranaction */
                        if (rdm_status == S_OKAY)
                        {
                            d_trend(task);
                        }
                        else
                        {
                            d_trabort(task);
                        }
                    }
                }
            }
        }
        else
        {
            /* Wait for a bit to allow the inserter thread to get more records
             * in the database
             */
            Sleep(500);
        }
    }

    /* Close the database and task */
    d_close(task);
    d_closetask(task);

    _endthreadex( 0 );
    return 0;
}

static uint32_t __stdcall ReaderThreadFcn(
    void *arg)
{
    DB_TASK  *task;
    int32_t   rdm_status;
    uint64_t  rec_count = 0;
    int32_t   field01_min;
    int32_t   field01_max;

    /* Create a task for this thread */
    rdm_status = d_opentask(&task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error open DB_TASK in ReaderThreadFcn - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        
		_endthreadex( rdm_status );
        return rdm_status;
    }

    /* Open the database in shared mode */
    rdm_status = d_open("core14_db", "s", task);
    if(rdm_status != S_OKAY)
    {
        fprintf(stderr, "Error opening database in ReaderThreadFcn - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status));
        d_closetask(task);
        
		_endthreadex( rdm_status );
        return rdm_status;
    }

    printf("\n");
    printf("Total Records   Min Val  Max Val\n");
    printf("-------------   -------  -------\n");

    /* Loop while we don't get an error or told to quit */
    while (!quit && rdm_status == S_OKAY)
    {
        /* Wait one second */
        Sleep(1000);

        /* Get information about the REC01 records */
        rdm_status = d_reclock(REC01, "r", task, CURR_DB);
        if(rdm_status == S_OKAY)
        {
            d_rectot(REC01, &rec_count, task, CURR_DB);
            d_keyfrst(REC01_FIELD01, task, CURR_DB);
            d_crread(REC01_FIELD01, &field01_min, task, CURR_DB);
            d_keylast(REC01_FIELD01, task, CURR_DB);
            d_crread(REC01_FIELD01, &field01_max, task, CURR_DB);            
            d_recfree(REC01, task, CURR_DB);
        }
        printf("       %6I64d    %6d   %6d\n", rec_count, field01_min, field01_max);
    }

    /* Close the database and task */
    d_close(task);
    d_closetask(task);

    _endthreadex( 0 );
    return 0;
}

/* =======================================================================
*/
int32_t main(void)
{
    DB_TASK      *task;
    int32_t       rdm_status;
    uint32_t      inserter_tid;
    uint32_t      deleter_tid;
    uint32_t      reader_tid;
    HANDLE        hInserterThread;
    HANDLE        hDeleterThread;
    HANDLE        hReaderThread;

    /* create the context for initializing the database */
    d_opentask(&task);

    /* Open the database for exclusive access, this will create the database
     * dictionary file (core14_db.dbd) if it does not already exist.
     */
    rdm_status = d_open_ptr("core14_db", "x", &core14_db_dbd, sizeof(core14_db_dbd), task);
    if (rdm_status != S_OKAY)
    {
        fprintf (stderr, "Exclusive open failed - %s: %s\n", d_errorname(rdm_status), d_errorstr(rdm_status) );
        d_closetask (task);
        return (1);
    }

    /* Initialize the databases.  Initializing will create the data files if 
     * they do not exist.  If the the data files do already exist they will be
     * removed and new empty files will be created.
     */
    d_initialize(task, CURR_DB);

    /* Close the database in the main thread */
    d_close(task);    

    printf ("\nStarting threads...\n");

    hInserterThread = (HANDLE) _beginthreadex(NULL, 0, InserterThreadFcn, NULL, 0, &inserter_tid);
    hDeleterThread  = (HANDLE) _beginthreadex(NULL, 0, DeleterThreadFcn, NULL, 0, &deleter_tid);
    hReaderThread   = (HANDLE) _beginthreadex(NULL, 0, ReaderThreadFcn, NULL, 0, &reader_tid);
    
    printf ("Press <Enter> when you are ready to terminate them . . .\n");
    getchar ();

    /* Tell the threads to quit */    
    quit = 1;

    /* Wait for the reader thread */
    WaitForSingleObject(hReaderThread, INFINITE);
    printf("Reader thread finished\n");
    CloseHandle(hReaderThread);

    /* Wait for the deleter thread */
    WaitForSingleObject(hDeleterThread, INFINITE);
    printf("Deleter thread finished\n");
    CloseHandle(hDeleterThread);

    /* Wait for the inserter thread */
    WaitForSingleObject(hInserterThread, INFINITE);
    printf("Inserter thread finished\n");
    CloseHandle(hInserterThread);

    d_closetask(task);

    printf ("\nPress <Enter> to continue . . .\n");
    getchar ();
    return 0;
}
